// RegistryPreferenceStore.cpp: implementation of the CRegistryPref class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "defines.h"
#include "RegistryPref.h"

using namespace PreferenceStore;


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CRegistryPref::CRegistryPref()
{

}

CRegistryPref::~CRegistryPref()
{

}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryPref::Load
// Description	    : 
// Return type		: bool 
// Argument         : const std::string & strPathName
//
bool CRegistryPref::Load( const std::string & strPathName )
{
	bool bRet(false);
	CRegistryKey theRootKey;
	if(theRootKey.Open(strPathName))
	{
		SKey * pRootKey;
		CreateKeys(pRootKey, strPathName, RootKey(), false);
		EnumSubKeys(theRootKey, pRootKey);
		theRootKey.Close();
		bRet = true;
	}
	return(bRet);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryPref::Save
// Description	    : 
// Return type		: bool 
// Argument         : const std::string & strPathName
//
bool CRegistryPref::Save( const std::string & strPathName )
{
	bool bRet(false);
	std::string strLastActiveKey;

	BuildPathToKey(strLastActiveKey, CurrentKey());

	strLastActiveKey = strLastActiveKey.substr(0, strLastActiveKey.find_last_of('\\'));

	if(SetKey(strPathName))
	{
		bRet = SaveTree(CurrentKey());
		SetKey(strLastActiveKey);
	}
	return(bRet);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: Save
// Description	    : 
// Return type		: bool 
// Argument         : const std::string & strPathName IPreferenceStore * pPreferences
//
bool CRegistryPref::Save( const std::string & strPathName, IPreferenceStore * pPreferences )
{
	bool bRet(false);

	std::string strLastActiveKey;
	pPreferences->GetKey(strLastActiveKey);

	if(pPreferences->SetKey(strPathName))
	{
		bRet = SaveTree(pPreferences->CurrentKey());
		pPreferences->SetKey(strLastActiveKey);
	}
	return(bRet);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryPref::EnumSubKeys
// Description	    : 
// Return type		: void 
// Argument         :  const CRegistryKey & refParentKey
// Argument         : SKey * pParentPreffKey
//
void CRegistryPref::EnumSubKeys( const CRegistryKey & refParentKey, PreferenceStore::SKey * pParentPreffKey )
{
	CRegistryKey::enum_context * pContext = NULL;
	CRegistryKey theSubKey;

	refParentKey.CreateEnumContext(pContext);

	EnumValues(refParentKey, pParentPreffKey);

	while(refParentKey.EnumSubKeys(theSubKey, pContext))
	{
		SKey *pKey;

		if(CreateKeys(pKey, theSubKey.Name(), *pParentPreffKey, false))
		{
			EnumSubKeys(theSubKey, pKey);
			theSubKey.Close();
		}
	}

	refParentKey.DestroyEnumContext(pContext);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryPref::EnumValues
// Description	    : 
// Return type		: void 
// Argument         :  const CRegistryKey & refKey
// Argument         : const CRegistryKey::enum_context * pContext
// Argument         : SKey * pParentPreffKey
//
void CRegistryPref::EnumValues( const CRegistryKey & refKey, PreferenceStore::SKey * pParentPreffKey )
{
	CRegistryKey::enum_context * pContext = NULL;
	CRegistryValue objRegValue;
	std::string strName;
	SPreferenceValue objValue;
	SPreferenceValue objValueName;

	refKey.CreateEnumContext(pContext);

	while(refKey.EnumValues(objRegValue, strName, pContext))
	{
		objRegValue.GetValue(objValue);
		objValue.SetDirty(false);
		pParentPreffKey->AddValue(objRegValue.GetName(), objValue, false);
	}

	refKey.DestroyEnumContext(pContext);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryPref::OnSaveKey
// Description	    : 
// Return type		: bool 
// Argument         : const std::string strPath
// Argument         : const SKey & refKey
//
bool CRegistryPref::OnSaveKey( const std::string strPath, const PreferenceStore::SKey & refKey ) const
{
	bool bRet = true;

	if(refKey.Dirty())
	{
		CRegistryKey theRegKey;
		if(refKey.Destroy())
		{
			bRet = theRegKey.Open(strPath);
			if(bRet)
			{
				bRet = theRegKey.Destroy();
			}
		}
		else
		{
			bRet = theRegKey.Create(strPath);
			if(bRet)
			{
				theRegKey.Close();
			}
		}	
	}

	return(bRet);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryPref::OnSaveValue
// Description	    : 
// Return type		: bool 
// Argument         : const std::string strPath
// Argument         : const SKey & refKey
// Argument         : const SPreferenceValue & refValueName
// Argument         : const SPreferenceValue & refValue
//
bool CRegistryPref::OnSaveValue( const std::string strPath, const PreferenceStore::SKey & refKey,
		const PreferenceStore::SPreferenceValue & refValueName, const PreferenceStore::SPreferenceValue & refValue ) const
{
	bool bRet = true;

	if (refValue.IsDirty())
	{
		CRegistryKey theRegKey;

		bRet = theRegKey.Open(strPath);
		if(bRet)
		{
			std::string strValueName;
			CRegistryValue theRegValue;

			refValueName.GetData(strValueName);
			if(refValue.Destroy())
			{
				bRet = theRegValue.Open(theRegKey, strValueName);
				if(bRet)
				{
					bRet = theRegValue.Destroy();
				}
			}
			else
			{
				bRet = theRegValue.Create(theRegKey, strValueName, refValue);
			}
		}
	}

	return(bRet);
}
