// WinRegistry.cpp: implementation of the CRegistryKey and CRegistryValue classes.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "defines.h"
#include "WinRegistry.h"

using namespace PreferenceStore;

IMPLEMENT_PREFIX_TABLE(CRegistryValue)
IMPLEMENT_POSTFIX_TABLE(CRegistryValue)

#define REG_SUCCESS(_arg_)  (ERROR_SUCCESS==(_arg_))


CRegistryKey::CRegKeyInfo::CRegKeyInfo()
	: m_dwSubKeys(0)
	, m_dwMaxSubKeyLen(0)
	, m_dwMaxClassLen(0)
	, m_dwValues(0)
	, m_dwMaxValueNameLen(0)
	, m_dwMaxValueLen(0)
	, m_dwSecurityDescriptor(0)
{
	memset(&m_ftLastWriteTime, 0, sizeof(m_ftLastWriteTime));
}

CRegistryKey::CRegKeyInfo::CRegKeyInfo(HKEY hKey)
{
	DWORD dwClassLen = 0;
	DWORD dwRet = RegQueryInfoKey(hKey, NULL, &dwClassLen, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

	if(ERROR_SUCCESS == dwRet || ERROR_MORE_DATA == dwRet)
	{
		dwClassLen++;
		m_strClass.reserve(dwClassLen);
		if(SUCCEEDED(RegQueryInfoKey(hKey, &m_strClass[0], &dwClassLen, NULL, &m_dwSubKeys, &m_dwMaxSubKeyLen,
						&m_dwMaxClassLen, &m_dwValues, &m_dwMaxValueNameLen, &m_dwMaxValueLen, &m_dwSecurityDescriptor,
						&m_ftLastWriteTime)))
		{
			m_dwMaxSubKeyLen++;
			m_dwMaxClassLen++;
			m_dwMaxValueNameLen++;
		}
	}
}

CRegistryKey::CRegKeyInfo::~CRegKeyInfo()
{
}

const CRegistryKey::CRegKeyInfo & CRegistryKey::CRegKeyInfo::operator = (const CRegKeyInfo & refToCopy)
{	
	m_strClass = refToCopy.m_strClass;
	m_dwSubKeys = refToCopy.m_dwSubKeys;
	m_dwMaxSubKeyLen = refToCopy.m_dwMaxSubKeyLen;
	m_dwMaxClassLen = refToCopy.m_dwMaxClassLen;
	m_dwValues = refToCopy.m_dwValues;
	m_dwMaxValueNameLen = refToCopy.m_dwMaxValueNameLen;
	m_dwMaxValueLen = refToCopy.m_dwMaxValueLen;
	m_dwSecurityDescriptor = refToCopy.m_dwSecurityDescriptor;
	memcpy(&m_ftLastWriteTime, &refToCopy.m_ftLastWriteTime, sizeof(FILETIME));
	return(*this);
}




CRegistryKey::CRegistryKey()
	: m_hKey(NULL)
{
}

CRegistryKey::~CRegistryKey()
{
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryKey::Open
// Description	    : 
// Return type		: bool 
// Argument         : const std::string & strPath
// Argument         : DWORD dwAccess
//
bool CRegistryKey::Open(const std::string & strPath, DWORD dwAccess) // dwAccess = KEY_ALL_ACCESS
{
	bool	bRet = false;
	HKEY	hKeyRoot;
	std::string	strSubKey;
	
	ExtractKeyHandleAndSubKey( hKeyRoot, strSubKey, strPath );

	if(Open(hKeyRoot, strSubKey, dwAccess))
	{
		m_theKeyInfo = CRegKeyInfo(m_hKey);
		m_strPath = strPath;
		ExtractKeyName(m_strKeyName, strPath);
		bRet = true;
	}
	return(bRet);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryKey::Close
// Description	    : 
// Return type		: void 
// Argument         : void
//
void CRegistryKey::Close()
{
	::RegCloseKey(m_hKey);
	m_hKey = NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryKey::Open
// Description	    : 
// Return type		: bool 
// Argument         : HKEY hKey
// Argument         : const std::string & strSubKey
// Argument         : DWORD dwAccess
//
bool CRegistryKey::Open(HKEY hKey, const std::string & strSubKey, DWORD dwAccess)
{
	return( REG_SUCCESS(::RegOpenKeyEx(hKey, strSubKey.c_str(), 0, dwAccess, &m_hKey)));
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryKey::Create
// Description	    : 
// Return type		: bool 
// Argument         : const std::string & strPath
// Argument         : DWORD dwAccess
// Argument         : DWORD dwOptions
//
bool CRegistryKey::Create(const std::string & strPath, DWORD dwAccess, DWORD dwOptions)
{
	bool	bRet = false;
	HKEY	hKeyRoot;
	std::string	strSubKey;

	ExtractKeyHandleAndSubKey( hKeyRoot, strSubKey, strPath );
	if(Create(hKeyRoot, strSubKey, dwAccess, dwOptions))
	{
		m_theKeyInfo = CRegKeyInfo(m_hKey);
		ExtractKeyName(m_strKeyName, strPath);
		m_strPath = strPath;
		bRet = true;
	}
	return(bRet);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryKey::Create
// Description	    : 
// Return type		: bool 
// Argument         : HKEY hKey
// Argument         : const std::string & strSubKey
// Argument         : DWORD dwAccess
// Argument         : DWORD dwOptions
//
bool CRegistryKey::Create(HKEY hKey, const std::string & strSubKey, DWORD dwAccess, DWORD dwOptions)
{
	DWORD dwDisposition;

	return(REG_SUCCESS(::RegCreateKeyEx(hKey, strSubKey.c_str(), 0, "", dwOptions, dwAccess, NULL, &m_hKey, &dwDisposition)));
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryKey::Destroy
// Description	    : 
// Return type		: bool 
// Argument         : void
//
bool CRegistryKey::Destroy( void )
{
	bool bRet = REG_SUCCESS(::RegDeleteKey(m_hKey, ""));
	m_hKey = NULL;
	return(bRet);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryKey::GetSubKey
// Description	    : 
// Return type		: bool 
// Argument         :  CRegistryKey & refKey
// Argument         : const std::string & strSubKey
// Argument         : DWORD dwAccess
//
bool CRegistryKey::GetSubKey( CRegistryKey & refKey, const std::string & strSubKey, DWORD dwAccess )
{
	bool bRet = refKey.Open(m_hKey, strSubKey, dwAccess);
	if(bRet)
	{
		refKey.m_theKeyInfo = CRegKeyInfo(m_hKey);
		refKey.m_strPath = m_strPath + '\\' + strSubKey;
		ExtractKeyName(refKey.m_strKeyName, strSubKey);
	}
	return(bRet);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryKey::EnumSubKeys
// Description	    : 
// Return type		: bool 
// Argument         :  CRegistryKey & refKey
// Argument         : enum_context *& pContext
// Argument         : DWORD dwAccess
//
bool CRegistryKey::EnumSubKeys( CRegistryKey & refKey, CRegistryKey::enum_context *& pContext, DWORD dwAccess ) const
{
	bool bRet(false);

	DWORD dwSubkeyLen = pContext->nBufferLen;
//	Initialize the key info and buffers

	DWORD dwRet = ::RegEnumKeyEx(m_hKey, pContext->dwCurrent, pContext->pszNameBuffer, &dwSubkeyLen, NULL, NULL, NULL, NULL);

	if(REG_SUCCESS(dwRet))
	{
		pContext->dwCurrent++;
		if(refKey.Open(m_hKey, pContext->pszNameBuffer, dwAccess))
		{
			ExtractKeyName(refKey.m_strKeyName, pContext->pszNameBuffer);
			refKey.m_strPath = m_strPath + '\\' + pContext->pszNameBuffer;
			refKey.m_theKeyInfo = CRegKeyInfo(refKey);
			bRet = true;
		}
//		else
//	Error exception
	}
//	else if(ERROR_NO_MORE_ITEMS != dwRet)
//	Error exception

	return(bRet);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryKey::EnumValues
// Description	    : 
// Return type		: bool 
// Argument         :  CRegistryValue & refValue
// Argument         : std::string & strName
// Argument         : enum_context *& pContext
//
bool CRegistryKey::EnumValues( CRegistryValue & refValue, std::string & strName, CRegistryKey::enum_context *& pContext ) const
{
	bool bRet = false;

	if(m_theKeyInfo.m_dwValues)
	{
		bRet = refValue.EnumValue(*this, pContext->dwCurrent, pContext);
	}
	return(bRet);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryKey::CreateEnumContext
// Description	    : 
// Return type		: void 
// Argument         : enum_context *& pContext
//
void CRegistryKey::CreateEnumContext( CRegistryKey::enum_context *& pContext ) const
{
	pContext = new enum_context;
	pContext->nBufferLen = m_theKeyInfo.m_dwMaxSubKeyLen > m_theKeyInfo.m_dwMaxValueNameLen ? m_theKeyInfo.m_dwMaxSubKeyLen : m_theKeyInfo.m_dwMaxValueNameLen;
	pContext->pszNameBuffer = new char[pContext->nBufferLen];
	pContext->dwCurrent = 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryKey::DestroyEnumContext
// Description	    : 
// Return type		: void 
// Argument         : enum_context *& pContext
//
void CRegistryKey::DestroyEnumContext( CRegistryKey::enum_context *& pContext ) const
{
	delete pContext->pszNameBuffer;
	delete pContext;
	pContext = NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryKey::ExtractKeyHandleAndSubKey
// Description	    : 
// Return type		: void 
// Argument         : HKEY & hKey
// Argument         : std::string & strSubKey
// Argument         : const std::string & strPath
//
void CRegistryKey::ExtractKeyHandleAndSubKey(HKEY & hKey, std::string & strSubKey, const std::string & strPath) const
{
	std::string m_strKeyName = strPath.substr(0, strPath.find('\\'));

	hKey = NULL;

	if(0 == _stricmp("HKEY_CLASSES_ROOT", m_strKeyName.c_str()))
		hKey = HKEY_CLASSES_ROOT;
	else if(0 == _stricmp("HKEY_CURRENT_CONFIG", m_strKeyName.c_str()))
		hKey = HKEY_CURRENT_CONFIG;
	else if(0 == _stricmp("HKEY_CURRENT_USER", m_strKeyName.c_str()))
		hKey = HKEY_CURRENT_USER;
	else if(0 == _stricmp("HKEY_LOCAL_MACHINE", m_strKeyName.c_str()))
		hKey = HKEY_LOCAL_MACHINE;
	else if(0 == _stricmp("HKEY_USERS", m_strKeyName.c_str()))
		hKey = HKEY_USERS;
	else if(0 == _stricmp("HKEY_PERFORMANCE_DATA", m_strKeyName.c_str()))
		hKey = HKEY_PERFORMANCE_DATA;
	else if(0 == _stricmp("HKEY_DYN_DATA", m_strKeyName.c_str()))
		hKey = HKEY_DYN_DATA;

	if(hKey && std::string::npos != strPath.find('\\'))
		strSubKey = strPath.substr(strPath.find('\\') + 1, strPath.size() - 1);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryKey::ExtractKeyName
// Description	    : 
// Return type		: void 
// Argument         : std::string & m_strKeyName
// Argument         : const std::string & strPath
//
void CRegistryKey::ExtractKeyName(std::string & m_strKeyName, const std::string & strPath) const
{
	m_strKeyName = strPath.substr(strPath.rfind('\\') + 1, strPath.size());
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: & CRegistryKey::operator = 
// Description	    : 
// Return type		: const CRegistryKey 
// Argument         : const CRegistryKey & refKeyToCopy
//
const CRegistryKey & CRegistryKey::operator = (const CRegistryKey & refKeyToCopy)
{
	m_hKey = refKeyToCopy.m_hKey;
	m_strKeyName = refKeyToCopy.m_strKeyName;
	m_strPath = refKeyToCopy.m_strPath;
	m_theKeyInfo = refKeyToCopy.m_theKeyInfo;
	return(*this);
}


///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryValue::Open
// Description	    : 
// Return type		: bool 
// Argument         : CRegistryKey & refKey
// Argument         : const std::string & strName
//
bool CRegistryValue::Open(CRegistryKey & refKey, const std::string & strName)
{
	bool bRet = false;
	m_theRegKey = refKey;
	m_strName = strName;
	DWORD dwType;
	DWORD dwValueLen = m_theRegKey.GetKeyInfo().m_dwMaxValueLen + 1;
	BYTE * pValue = new BYTE[dwValueLen];

	if(REG_SUCCESS(::RegQueryValueEx(m_theRegKey, m_strName.c_str(), 0, &dwType, pValue, &dwValueLen)))
	{
		switch(dwType)
		{
		case REG_SZ:
			m_theValue.SetData(std::string((const char*)pValue));
			bRet = true;
			break;
		case REG_DWORD:
			m_theValue.SetData((uint32)*(DWORD*)pValue);
			bRet = true;
			break;
		case REG_BINARY:
			m_theValue.SetData(pValue, dwValueLen);
			bRet = true;
			break;
		default:
//	Error exception
			break;
		}
	}
	return(bRet);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryValue::Create
// Description	    : 
// Return type		: bool 
// Argument         : CRegistryKey & refKey
// Argument         : const std::string & strName
// Argument         : const SPreferenceValue & refValue
//
bool CRegistryValue::Create(CRegistryKey & refKey, const std::string & strName, const SPreferenceValue & refValue)
{
	bool bRet = false;
	m_theRegKey = refKey;
	m_strName = strName;
	if(SetValue(refValue))
	{
		bRet = true;
	}
	else
	{
		m_strName.erase();
	}
	return(bRet);
}

bool CRegistryValue::EnumValue(const CRegistryKey & refKey, size_t nIndex, CRegistryKey::enum_context * pContext /*= NULL*/)
{
	bool bRet = false;
	DWORD dwNameLen;
	char * pszNameBuffer;
	DWORD dwDataLen = refKey.GetKeyInfo().m_dwMaxValueLen;
	DWORD dwType;

	if(pContext)
	{
		dwNameLen = pContext->nBufferLen;
		pszNameBuffer = pContext->pszNameBuffer;
	}
	else
	{
		dwNameLen = refKey.GetKeyInfo().m_dwMaxValueNameLen;
		pszNameBuffer = new char[dwNameLen];
	}

	BYTE * pData = new BYTE[dwDataLen];

	if(REG_SUCCESS(::RegEnumValue(refKey, static_cast<DWORD>(nIndex), pszNameBuffer, &dwNameLen, NULL, &dwType, pData, &dwDataLen)))
	{
		pContext->dwCurrent++;
		m_strName = pszNameBuffer;
		switch(dwType)
		{
		case REG_BINARY:
			m_theValue.SetData((void*)pData, dwDataLen);
			bRet = true;
			break;
		case REG_DWORD:
			m_theValue.SetData((uint32)*((DWORD*)pData));
			bRet = true;
			break;
		case REG_SZ:
			ConvertToValue(m_theValue, (char*)pData);
			bRet = true;
			break;
		default:
//	Error exception
			break;
		}
	}
	delete pData;

	return(bRet);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryValue::Set
// Description	    : 
// Return type		: bool 
// Argument         : const SPreferenceValue & refValue
//
bool CRegistryValue::SetValue(const SPreferenceValue & refValue)
{
	bool bRet = false;

	m_theValue = refValue;
	switch(refValue.GetType())
	{
	case SPreferenceValue::kStringType:
		{
			std::string strValue;
			refValue.GetData(strValue);

			bRet = REG_SUCCESS(::RegSetValueEx(m_theRegKey, m_strName.c_str(), 0,
				REG_SZ, (BYTE*)strValue.c_str(), static_cast<DWORD>(strValue.size())));
		}
		break;
	case SPreferenceValue::kUnsignedIntegerType:
		{
			DWORD dwValue;
			refValue.GetData((uint32&)dwValue);

			bRet = REG_SUCCESS(::RegSetValueEx(m_theRegKey, m_strName.c_str(), 0,
				REG_DWORD, (BYTE*)&dwValue, sizeof(DWORD)));
		}
		break;
	case SPreferenceValue::kBinaryType:
		{
			size_t nDataLen = refValue.GetSize();
			BYTE * pData = new BYTE[nDataLen];
			refValue.GetData(pData, nDataLen);

			bRet = REG_SUCCESS(::RegSetValueEx(m_theRegKey, m_strName.c_str(), 0,
				REG_BINARY, pData, static_cast<DWORD>(nDataLen) ));
			delete pData;
		}
		break;
	case SPreferenceValue::kSignedIntegerType:
	case SPreferenceValue::kFloatType:
	case SPreferenceValue::kVector3Type:
	case SPreferenceValue::kVector4Type:
	case SPreferenceValue::kRGBAColorType:
		{
			std::string strValue = refValue.GetDataAsText();

			strValue = std::string(GET_PREFIX(refValue.GetType())) +
				strValue + GET_POSTFIX(refValue.GetType());

			bRet = REG_SUCCESS(::RegSetValueEx(m_theRegKey, m_strName.c_str(), 0,
				REG_SZ, (BYTE*)strValue.c_str(), static_cast<DWORD>(strValue.size())));
		}
		break;
	}
	return(bRet);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CRegistryValue::Destroy
// Description	    : 
// Return type		: bool 
// Argument         : void
//
bool CRegistryValue::Destroy(void)
{
	bool bRet = REG_SUCCESS(::RegDeleteValue(m_theRegKey, m_strName.c_str()));
	m_strName.erase();
	return(bRet);
}

///////////////////////////////////////////////////////////////////////////////
// Function name   : CRegistryValue::ConvertToValue
// Description     : 
// Return type     : void 
// Argument        :  SPreferenceValue & refValue
// Argument        : const std::string & strData
//
void CRegistryValue::ConvertToValue( SPreferenceValue & refValue, const std::string & strData ) const
{
	static std::string strBuffer;

	if(!strData.empty())
	{
		if(strData[0] == '\"')
		{
			strBuffer = strData;
			Trim(strBuffer, "\"");
			refValue.SetData(strBuffer);
		}
		else
		{
			char		*pBuffer = new char[strData.size() + 1];
			char		*pTemp;
			std::string	strToken;
			std::string	strValue;
			SPreferenceValue::ValueType	eType = SPreferenceValue::kUndefinedType;

			strcpy(pBuffer, strData.c_str());

			pTemp = strtok(pBuffer, ":");
			_strlwr(pTemp);
			strToken = pTemp;
			strToken += ':';

			if(0 == strncmp(strToken.c_str(), GET_PREFIX(SPreferenceValue::kStringType), strToken.size()))
				eType = SPreferenceValue::kStringType;
			else if(0 == strcmp(strToken.c_str(), GET_PREFIX(SPreferenceValue::kUnsignedIntegerType)))
				eType = SPreferenceValue::kUnsignedIntegerType;
			else if(0 == strcmp(strToken.c_str(), GET_PREFIX(SPreferenceValue::kBinaryType)))
				eType = SPreferenceValue::kBinaryType;
			else if(0 == strcmp(strToken.c_str(), GET_PREFIX(SPreferenceValue::kSignedIntegerType)))
				eType = SPreferenceValue::kSignedIntegerType;
			else if(0 == strcmp(strToken.c_str(), GET_PREFIX(SPreferenceValue::kFloatType)))
				eType = SPreferenceValue::kFloatType;
			else if(0 == strcmp(strToken.c_str(), GET_PREFIX(SPreferenceValue::kVector3Type)))
				eType = SPreferenceValue::kVector3Type;
			else if(0 == strcmp(strToken.c_str(), GET_PREFIX(SPreferenceValue::kVector4Type)))
				eType = SPreferenceValue::kVector4Type;
			else if(0 == strcmp(strToken.c_str(), GET_PREFIX(SPreferenceValue::kRGBAColorType)))
				eType = SPreferenceValue::kRGBAColorType;
			
			if(eType == SPreferenceValue::kUndefinedType)
			{
				eType = SPreferenceValue::kStringType;
				strValue = strData.c_str();
			}
			else
			{
				strValue = strData.substr(strToken.size());
			}


			if(eType == SPreferenceValue::kStringType)
				Trim(strValue, "\"");

			refValue.SetRawData(strValue.empty() ? "" : strValue, eType);
			delete[] pBuffer;
		}
	}
	refValue.SetDirty(false);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function name	: CPreferenceStore::Trim
// Description	    : Removes any leading and trailing values in strRemove from strToTrim
// Return type		: void 
// Argument         : std::string & strToTrim
// Argument         : const std::string & strRemove
//
void CRegistryValue::Trim( std::string & strToTrim, const std::string & strRemove /*= " "*/ ) const
{
	std::string::size_type nBegin = strToTrim.find_first_not_of(strRemove);
	std::string::size_type nEnd = strToTrim.find_last_not_of(strRemove);
	if(std::string::npos == nBegin)
	{
		strToTrim.erase();
	}
	else
	{
		nEnd = std::string::npos == nEnd ? 0 : nEnd;
		strToTrim = strToTrim.substr( nBegin, nEnd - nBegin + 1);
	}
}
